from .burning_ship import BurningShip
from .inv_mandel import InverseMandelbrot
from .julia import Julia
from .mandelbrot import Mandelbrot

fractal_list = [Mandelbrot, BurningShip, Julia, InverseMandelbrot]
